from contextlib import contextmanager
from pathlib import Path
from typing import Any, Callable, Optional, Union
from contextvars import copy_context
from instaui.runtime import new_app_slot, reset_app_slot
from instaui.systems import file_path_system
from instaui.runtime.utils import init_base_scope
from .func import to_html, to_html_str, get_template_model
from instaui.template.zero_template import ZeroTemplateModel
from instaui.launch_collector import get_launch_collector


class ZeroCaller:
    def __init__(
        self,
        *,
        icons_svg_path: Optional[Union[str, Path]] = None,
    ):
        self._mate = {}

        if icons_svg_path is not None:
            icons_svg_path = file_path_system.get_caller_path(icons_svg_path)
            self._mate["icons_svg_code"] = icons_svg_path.read_text(encoding="utf-8")

    def to_html(self, render_fn: Callable[..., Any], *, file: Union[str, Path]):
        file = file_path_system.get_caller_path(file)

        with _run(self._mate):
            copy_context().run(render_fn)
            return to_html(file)

    def to_html_str(self, render_fn: Callable[..., Any]):
        with _run(self._mate):
            copy_context().run(render_fn)
            return to_html_str()

    def to_debug_report(self, render_fn: Callable[..., Any], *, file: Union[str, Path]):
        file = file_path_system.get_caller_path(file)

        # Custom component dependencies must be recorded only during actual execution
        with _run(self._mate):
            copy_context().run(render_fn)
            result_html_str = to_html_str()

        model = get_template_model()
        with _run({"merge_global_resources": False}):
            _create_debug_report(model, result_html_str)
            to_html(file.resolve().absolute())


@contextmanager
def _run(mate: dict):
    app, app_token = new_app_slot("zero", app_meta=mate)
    init_base_scope(app)

    _events = [iter(event()) for event in get_launch_collector().page_init_lifespans]
    for event in _events:
        next(event)

    yield

    for event in _events:
        try:
            next(event)
        except StopIteration:
            pass

    assert app_token is not None
    reset_app_slot(app_token)


def _create_debug_report(model: ZeroTemplateModel, result_html_str: str):
    from instaui import ui, html

    no_exists_path_class = "ex-no-exists-path"

    def _path_exists_class(path: Path):
        return "" if path.exists() else no_exists_path_class

    ui.use_tailwind()

    ui.add_style(rf".{no_exists_path_class} {{background-color: red;color: white;}}")

    html_size = len(result_html_str.encode("utf-8")) / 1024 / 1024

    box_style = "border-2 border-gray-200 p-4 place-center gap-x-2"

    with ui.column().classes("gap-2"):
        # base info
        with ui.grid(columns="auto 1fr").classes(box_style):
            html.span("file size:")
            html.span(f"{html_size:.2f} MB")

        # import maps
        html.paragraph("import maps")

        rows = [
            ["vue", str(model.vue_js_code)],
            ["instaui", str(model.instaui_js_code)],
        ]

        for name, url in model.extra_import_maps.items():
            if isinstance(url, Path) and url.is_file():
                rows.append([name, str(url)])

        html.table(["name", "path"], rows).scoped_style(
            r"""
    table {
    outline: 1px solid black;
    }
    td, th {
    outline: 1px solid black;
    padding: 5px;
    }
    """,
            with_self=True,
        )

        # css links
        html.paragraph("css links")
        with ui.column().classes(box_style):
            for link in model.css_links:
                if isinstance(link, Path) and link.is_file():
                    html.span(str(link)).classes(_path_exists_class(link))

        # js links
        html.paragraph("js links")
        with ui.column().classes(box_style):
            for info in model.js_links:
                if isinstance(info.link, Path) and info.link.is_file():
                    html.span(str(info.link)).classes(_path_exists_class(info.link))

        # custom components
        html.paragraph("custom components")
        with ui.grid(columns="auto 1fr").classes(box_style):
            html.span("name")
            html.span("js file path")

            for info in model.vue_app_component:
                html.span(info.name)

                if isinstance(info.url, Path) and info.url.is_file():
                    html.span(str(info.url)).classes(_path_exists_class(info.url))
                else:
                    html.span("not file")



from __future__ import annotations
from instaui.common.binding_track_mixin import try_mark_binding
from instaui.common.component_mixin import ComponentMixin
from instaui.common.var_track_mixin import mark_as_used
from instaui.vars.types import TMaybeRef
from instaui.components.logic_component import LogicContainer


class VIf(LogicContainer):
    def __init__(self, on: TMaybeRef[bool]):
        super().__init__("vif")
        mark_as_used(on)
        self._on = try_mark_binding(on)
        self._items: list[ComponentMixin] = []

    def _to_json_dict(self):
        data = super()._to_json_dict()

        data["on"] = self._on

        if self._items:
            data["items"] = self._items

        return data

    def add_item(self, item: ComponentMixin):
        self._items.append(item)



from typing import Tuple, TypeVar, cast
from instaui.runtime.scope import Scope
from instaui.vars.path_var import PathVar
from instaui.vars.ref import Ref
from instaui.vars._types import InputBindingType, OutputSetType
from instaui.vars.mixin_types.py_binding import CanInputMixin, CanOutputMixin
from instaui.vars.mixin_types.observable import ObservableMixin
from instaui.vars.mixin_types.element_binding import ElementBindingMixin
from instaui.vars.mixin_types.str_format_binding import StrFormatBindingMixin
from instaui.common.var_track_mixin import VarTrackerMixin
from instaui.common.binding_track_mixin import BindingTrackerMixin

from instaui.common.jsonable import Jsonable
from instaui.vars.ref_base import RefBase


from pydantic import BaseModel, RootModel


_T = TypeVar("_T")


_ProxyModel = RootModel


class RefProxy(
    CanInputMixin,
    ObservableMixin,
    CanOutputMixin,
    StrFormatBindingMixin,
    ElementBindingMixin,
    VarTrackerMixin,
    BindingTrackerMixin,
    PathVar,
    Jsonable,
):
    def __init__(self, data, ref_base: RefBase) -> None:
        self._ref_ = ref_base
        self._prop_names_ = set(data.keys()) if isinstance(data, dict) else set()

    @property
    def __ref_(self):
        return super().__getattribute__("_ref_")

    def __getattribute__(self, name):
        if name not in super().__getattribute__("_prop_names_"):
            return super().__getattribute__(name)

        return self.__ref_[name]

    def __getitem__(self, name):
        return self.__ref_[name]

    def not_(self):
        return self.__ref_.not_()

    def __add__(self, other: str):
        return self.__ref_ + other

    def __lt__(self, other):
        return self.__ref_ < other

    def __le__(self, other):
        return self.__ref_ <= other

    def __gt__(self, other):
        return self.__ref_ > other

    def __ge__(self, other):
        return self.__ref_ >= other

    def __ne__(self, other):
        return self.__ref_ != other

    def len_(self):
        return self.__ref_.len_()

    def __radd__(self, other: str):
        return other + self.__ref_

    def _to_str_format_binding(self, order: int) -> Tuple[str, str]:
        return self.__ref_._to_str_format_binding(order)

    def _to_json_dict(self):
        return self.__ref_._to_json_dict()

    def _to_event_output_type(self) -> OutputSetType:
        return self.__ref_._to_event_output_type()

    def _to_event_input_type(self) -> InputBindingType:
        return self.__ref_._to_event_input_type()

    def _mark_as_used(self):
        self._ref_._mark_as_used()

    def _mark_binding(self, scope: Scope) -> dict:
        return self._ref_._mark_binding(scope)


class StateModel(BaseModel, Jsonable):
    pass

    def _to_json_dict(self):
        return self.model_dump()


def state(
    value: _T,
    deep_compare: bool = False,
) -> _T:
    """
    Creates a reactive state object that tracks changes and notifies dependencies.

    Args:
        value (_T): The initial value to wrap in a reactive state container.
                   Can be any type (primitive, object, or complex data structure).

    # Example:
    .. code-block:: python
        from instaui import ui,html
        count = ui.state(0)

        html.number(count)
        ui.text(count)
    """

    if isinstance(value, RefProxy):
        return value

    new_value = _ProxyModel(value).model_dump()
    obj = RefProxy(new_value, Ref(value, deep_compare=deep_compare))
    return cast(_T, obj)



from typing import ClassVar
from typing_extensions import Self
from contextvars import ContextVar


class PageState:
    _ctx: ClassVar[ContextVar["PageState"]]

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls._ctx = ContextVar(f"{cls.__name__}_ctx")

    @classmethod
    def get(cls) -> Self:
        inst = cls._ctx.get(None)

        if inst is None:
            inst = cls()
            cls._ctx.set(inst)
        return inst  # type: ignore



from __future__ import annotations
from typing import Optional, Literal
from typing_extensions import Unpack
from instaui.components.element import Element
from instaui.vars.types import TMaybeRef
from .base_props import TLayoutBaseProps
from instaui.components._responsive_type._common import TMaybeResponsive
from instaui.custom import configure_slot_without_slot_prop


class Box(Element):
    def __init__(
        self,
        *,
        as_: Optional[TMaybeRef[Literal["div", "span"]]] = None,
        as_child: Optional[TMaybeRef[bool]] = None,
        display: Optional[
            TMaybeResponsive[Literal["none", "inline-block", "block", "contents"]]
        ] = None,
        **kwargs: Unpack[TLayoutBaseProps],
    ):
        super().__init__("box")
        configure_slot_without_slot_prop(self)

        self.props(
            {
                "as": as_,
                "as_child": as_child,
                "display": display,
                **kwargs,
            }
        )



import typing
from instaui.common.binding_track_mixin import (
    BindingTrackerMixin,
    try_mark_binding,
    is_binding_tracker,
)
from instaui.common.var_track_mixin import VarTrackerMixin, mark_as_used
from instaui.runtime.scope import Scope
from instaui.vars._types import InputBindingType
from instaui.vars.mixin_types.py_binding import CanInputMixin


class InputSilentData(CanInputMixin, BindingTrackerMixin, VarTrackerMixin):
    def __init__(self, value: typing.Union[BindingTrackerMixin, typing.Any]) -> None:
        self.value = value

    def is_const_value(self) -> bool:
        return not is_binding_tracker(self.value)

    def _to_event_input_type(self) -> InputBindingType:
        return InputBindingType.Ref

    def _mark_binding(self, scope: Scope) -> dict:
        return try_mark_binding(self.value, scope=scope)

    def _mark_as_used(self):
        mark_as_used(self.value)



from __future__ import annotations
from typing import Literal, Union
from typing_extensions import Unpack
from instaui.components.element import Element
from instaui.custom import configure_slot_without_slot_prop
from instaui.vars.types import TMaybeRef
from .base_props import TLayoutBaseProps
from instaui.components._responsive_type._common import TMaybeResponsive, TLevel_0_9


class TFlexBaseProps(TLayoutBaseProps, total=False):
    as_: TMaybeRef[Literal["div", "span"]]
    as_child: TMaybeRef[bool]
    display: TMaybeResponsive[Literal["none", "inline-flex", "flex"]]
    align: TMaybeResponsive[Literal["start", "center", "end", "stretch", "baseline"]]
    justify: TMaybeResponsive[Literal["start", "center", "end", "between"]]
    wrap: TMaybeResponsive[Literal["nowrap", "wrap", "wrap-reverse"]]
    gap: TMaybeResponsive[Union[str, TLevel_0_9]]
    gap_x: TMaybeResponsive[Union[str, TLevel_0_9]]
    gap_y: TMaybeResponsive[Union[str, TLevel_0_9]]


class TFlexProps(TFlexBaseProps, total=False):
    direction: TMaybeResponsive[
        Literal["row", "column", "row-reverse", "column-reverse"]
    ]


class Flex(Element):
    def __init__(
        self,
        **kwargs: Unpack[TFlexProps],
    ):
        super().__init__("flex")
        configure_slot_without_slot_prop(self)

        self.props(kwargs)


class FlexRow(Flex):
    def __init__(
        self,
        **kwargs: Unpack[TFlexBaseProps],
    ):
        super().__init__(direction="row", **kwargs)


class FlexColumn(Flex):
    def __init__(
        self,
        **kwargs: Unpack[TFlexBaseProps],
    ):
        super().__init__(direction="column", **kwargs)
